﻿-- Initialize Per Character values
settings = {
	general = {
		allowDrag = true,
		allowSettingsClick = true,
		useEnglish = false
	},
	messages = {
		playerFloating = true,
		playerChat = false,
		petFloating = true,
		petChat = false
	},
	averageDisplay = {
		visible = true,
		backdrop = true,
		showPlayerKills = true,
		showPlayerQuests = true,
		showPetKills = true
	},
}
data = {
	player = {
		killAverage = nil,
		questAverage = nil
	},
	pet = {
		killAverage = nil
	}
}

-- local values
local maxLevel = 80;
local currentLevel, xpTotal, xpCurrent, xpRemaining, xpRested
local lastMobName, lastMobXP, lastQuestXP
local killsRemaining, questsRemaining
local mobAverageTable = {}
local questAverageTable = {}
local locale = nil
local displayLocale = nil
local backdropTable = nil

-- Waterfall GUI Config
local waterfall = AceLibrary("Waterfall-1.0")

----------------------
-- Initialization
----------------------
-- Main entry point
function MainOnload()
	-- Disable if the character is at max level
	local currentLevel = UnitLevel("player")
	
	if(currentLevel < maxLevel) then
		-- Get the locale
		locale = GetLocale();
		displayLocale = locale; -- Chosen locale loaded in the onPlayerLogin callback.
		
		-- Check if the client language is suported
		if(LOCALE_SUPPORTED[locale] ~= true) then
			if LOCALE_SUPPORTED[locale] == false then
				DEFAULT_CHAT_FRAME:AddMessage("XToLevel could not be loaded. Client language is not supported.");
			else
				DEFAULT_CHAT_FRAME:AddMessage(LOCALE_SUPPORTED[locale]);
			end
			return;
		end
		
		-- Show the frame
		XToLevelDisplay:Show();
		
		-- Register Events
		this:RegisterEvent("CHAT_MSG_COMBAT_XP_GAIN");
		this:RegisterEvent("PLAYER_LOGIN");
		this:RegisterEvent("PLAYER_LEVEL_UP");
		
		-- Register slash commands
		SLASH_XTOLEVEL1 = "/xtolevel";
		SLASH_XTOLEVEL2 = "/xtl";
		SlashCmdList["XTOLEVEL"] = SlashHandler;
	end
end
function SlashHandler()
	waterfall:Open('ConfigWindow', 'general')
end
function ConfigWindow_Init()
	waterfall:Register('ConfigWindow',
		'title',LOCALE_CONFIG_TEXT[displayLocale]["Title"],
		'colorR', 0.0,
		'colorG', 0.5,
		'colroB', 0.25,
		'tree', 
		{
			{
				text = LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["General"]["Name"],
				id = 'general',
				isOpen = true,
				isSelected = true
			},
			{
				text = LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["Messages"]["Name"],
				id = 'messages',
				isOpen = true,
				isSelected = true
			},
			{
				text = LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["AverageDisplay"]["Name"],
				id = 'average',
				isOpen = true,
				isSelected = true
			}
		},
	 'children', 
		function(id)
			if id=="general" then
				if locale ~= "enUS" then
					waterfall:AddControl(
						'type', 'checkbox',
						'text', LOCALE_CONFIG_TEXT[locale]["Categories"]["General"]["UseEnglish"],
						'checked', settings["general"]["useEnglish"],
						'fullRefresh', true,
						'setFunc', function() ToogleUseEnglish() end
					)
				else
					waterfall:AddControl(
					'type', 'label',
					'text', "No options available"
		        )
				end
				
				return LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["Messages"]["Name"]
			elseif id=='messages' then
				waterfall:AddControl(
					'type', 'label',
					'text', LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["Messages"]["PlayerMessagesLabel"]
		        )
				waterfall:AddControl(
					'type', 'checkbox',
					'text', LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["Messages"]["ShowInChat"],
					'checked', settings['messages']['playerChat'],
					'fullRefresh', true,
					'setFunc', function() settings['messages']['playerChat'] = not settings['messages']['playerChat'] end
		        )
				waterfall:AddControl(
					'type', 'checkbox',
					'text', LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["Messages"]["ShowFloating"],
					'checked', settings['messages']['playerFloating'],
					'fullRefresh', true,
					'setFunc', function() settings['messages']['playerFloating'] = not settings['messages']['playerFloating'] end
		        )
				
				return LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["Messages"]["Name"]
			elseif id=='average' then
				waterfall:AddControl(
					'type', 'label',
					'text', LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["AverageDisplay"]["BehaviorLabel"]
		        )
				waterfall:AddControl(
					'type', 'checkbox',
					'text', LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["AverageDisplay"]["LockAverageDisplay"],
					'checked', not settings['general']['allowDrag'],
					'fullRefresh', true,
					'setFunc', function() ToogleAverageDisplayLock() end
		        )
				waterfall:AddControl(
					'type', 'checkbox',
					'text', LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["AverageDisplay"]["AllowSettingsClick"],
					'checked', settings['general']['allowSettingsClick'],
					'fullRefresh', true,
					'setFunc', function() ToogleSettingsClick() end
		        )
			
				waterfall:AddControl(
					'type', 'label',
					'text', ""
		        )
				
				waterfall:AddControl(
					'type', 'label',
					'text', LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["AverageDisplay"]["AppearenceLabel"]
		        )
				waterfall:AddControl(
					'type', 'checkbox',
					'text', LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["AverageDisplay"]["ShowAverageDisplay"],
					'checked', settings['averageDisplay']['visible'],
					'fullRefresh', true,
					'setFunc', function() ToogleDisplay() end
		        )
				waterfall:AddControl(
					'type', 'checkbox',
					'text', LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["AverageDisplay"]["ShowFrame"],
					'checked', settings['averageDisplay']['backdrop'],
					'fullRefresh', true,
					'setFunc', function() ToogleBackdrop() end
		        )
				
				waterfall:AddControl(
					'type', 'label',
					'text', ""
		        )
				
				waterfall:AddControl(
					'type', 'label',
					'text', LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["AverageDisplay"]["DataLabel"]
		        )
				waterfall:AddControl(
					'type', 'checkbox',
					'text', LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["AverageDisplay"]["ShowKills"],
					'checked', settings['averageDisplay']['showPlayerKills'],
					'fullRefresh', true,
					'setFunc', function() ToogleMobDisplay() end
		        )
				waterfall:AddControl(
					'type', 'checkbox',
					'text', LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["AverageDisplay"]["ShowQuests"],
					'checked', settings['averageDisplay']['showPlayerQuests'],
					'fullRefresh', true,
					'setFunc', function() ToogleQuestDisplay() end
		        )
				return LOCALE_CONFIG_TEXT[displayLocale]["Categories"]["AverageDisplay"]["Name"]
			end
		end 
	)
end

----------------------
-- GUI Config callbacks
----------------------
function ToogleBackdrop()
	settings['averageDisplay']['backdrop'] = not settings['averageDisplay']['backdrop'];
	UpdateAverageBackdrop();
end
function ToogleDisplay()
	settings['averageDisplay']['visible'] = not settings['averageDisplay']['visible'];
	if settings['averageDisplay']['visible'] then
		XToLevelDisplay:Show();
		UpdateAverageDisplay();
	else
		XToLevelDisplay:Hide();
	end
end
function ToogleMobDisplay()
	settings['averageDisplay']['showPlayerKills'] = not settings['averageDisplay']['showPlayerKills'];
	if settings['averageDisplay']['visible'] then
		XToLevelDisplay:Show();
		UpdateAverageDisplay();
	end
end
function ToogleQuestDisplay()
	settings['averageDisplay']['showPlayerQuests'] = not settings['averageDisplay']['showPlayerQuests'];
	if settings['averageDisplay']['visible'] then
		XToLevelDisplay:Show();
		UpdateAverageDisplay();
	end
end
function ToogleUseEnglish()
	settings["general"]["useEnglish"] = not settings['general']['useEnglish'];
	if settings['general']['useEnglish'] then
		displayLocale = "enUS";
	else
		displayLocale = locale;
	end
	UpdateAverageDisplay();
	waterfall:Close('ConfigWindow');
	ConfigWindow_Init();
	waterfall:Open('ConfigWindow', 'general');
end
function ToogleSettingsClick()
	settings["general"]["allowSettingsClick"] = not settings["general"]["allowSettingsClick"];
end
function ToogleAverageDisplayLock()
	settings["general"]["allowDrag"] = not settings["general"]["allowDrag"]
end

----------------------
-- Display functions
----------------------
function SetTextSpot(spotNumber, newText)
	if spotNumber == 1 then
		XToLevelDisplayTextSpot1:SetText(newText);
		return XToLevelDisplayTextSpot1:GetWidth();
	elseif spotNumber == 2 then
		XToLevelDisplayTextSpot2:SetText(newText);
		return XToLevelDisplayTextSpot2:GetWidth();
	end
end
function UpdateAverageBackdrop()
	if settings['averageDisplay']['visible'] then
		XToLevelDisplayLabelText:SetText("XToLevel");
		if settings['averageDisplay']['backdrop'] then
			XToLevelDisplay:SetBackdrop(backdropTable);
			XToLevelDisplay:SetBackdropColor("0.0", "0.75", "0.5", "0.75");
			XToLevelDisplay:SetBackdropBorderColor("0.0", "0.0", "0.0", "1.0");
		else
			XToLevelDisplay:SetBackdrop(nil);
		end
	end
end
function UpdateAverageDisplay()
	XToLevelDisplayLabelText:SetText("");
	XToLevelDisplayTextSpot1:SetText("");
	XToLevelDisplayTextSpot2:SetText("");
	if settings['averageDisplay']['visible'] then
		XToLevelDisplayLabelText:SetText("XToLevel");
		local spotNumber = 0;
		local maxWidth = XToLevelDisplayLabelText:GetWidth();
		local currentWidth = 0;
		if killsRemaining ~= nil and settings['averageDisplay']['showPlayerKills'] == true then
			spotNumber = spotNumber + 1
			currentWidth = 5 + SetTextSpot(spotNumber, LOCALE_MESSAGES[displayLocale]['Kills'] ..": ".. killsRemaining)
			if maxWidth < currentWidth then
				maxWidth = currentWidth;
			end
		end
		if questsRemaining ~= nil and settings['averageDisplay']['showPlayerQuests'] then
			spotNumber = spotNumber + 1
			currentWidth = 5 + SetTextSpot(spotNumber, LOCALE_MESSAGES[displayLocale]['Quests'] ..": ".. questsRemaining)
			if maxWidth < currentWidth then
				maxWidth = currentWidth;
			end
		end
		XToLevelDisplay:SetHeight(28 + (spotNumber * 12));
		XToLevelDisplay:SetWidth(maxWidth + 17);
	end
end
function ShowPlayerMessages(text)
	if settings['messages']['playerChat'] then
		DEFAULT_CHAT_FRAME:AddMessage(text, 0, 100, 200, 55, 4);
	end
	if settings['messages']['playerFloating'] then
		UIErrorsFrame:AddMessage(text, 0, 100, 200, 55, 4);
	end
end


----------------------
-- Misc functions
----------------------
-- Used to find the Zero Difference used to calculate XP from low level mobs
function GetZeroDifference(charLevel)
	for i,v in ipairs(ZD_Table) do
		if charLevel >= v[1] and charLevel <= v[2] then
			return v[3];
		end
	end
end
-- Returns the ID of the zone the player is current in
function GetZoneID()
	local currentZone = GetRealZoneText();
	local continentNames, key, val = { GetMapContinents() } ;
	for key, val in pairs(continentNames) do
		local continentZones = { GetMapZones(key) };
		for i,zone in ipairs(continentZones) do
			if zone == currentZone then
				return i
			end
		end
	end
	return nil
end
-- Extracts the mob name and XP gained from the combat XP message
function ExtractKillInfoReg(message)
	local pattern, mob, xp;
	--- Check if this is a quest or a mob
	if(string.find(message, ",") == nil) then
		pattern = LOCALE_XPGAIN_MSG[locale];
		xp = strmatch(message, pattern);
	else
		pattern = LOCALE_KILL_MSG[locale];
		mob, xp = strmatch(message, pattern);
	end
	
	return mob, xp;
end

----------------------
-- Experience calculations
----------------------
-- Get total rested bonus available for this level
function GetRestedToLevel()
	local xpToLevel = xpTotal - xpCurrent;
	local restedToLevel = 0;
	if(xpRested == 0) then
		restedToLevel = 0;
	elseif(xpToLevel / 2 < xpRested) then
		restedToLevel = xpToLevel / 2;
	else
		restedToLevel = xpRested;
	end
	return restedToLevel
end
-- Gets the total XP for the given level
function GetXPMax(level)
	local unit = UnitXPMax("player")
	if(level < 70 and unit < XP_TABLE[level][3]) then
		return XP_TABLE[level][3]
	else
		return unit
	end
end
-- Calculates the XP gained from killing a mob
function GetMobXP(charLevel, mobLevel)
	-- Make sure the charLevel is set, to avoid null reference errors
	if charLevel == nil then
		charLevel = UnitLevel("player");
	end
	if mobLevel == nil then 
		mobLevel = charLevel;
	end

	-- Find the gray level
	local grayLevel = 0
	if charLevel < 6 then
		grayLevel = 0
	elseif charLevel < 40 then
		grayLevel = charLevel - floor(charLevel / 10) - 5
	elseif charLevel < 60 then
		grayLevel = charLevel - floor(charLevel / 5) - 1
	else
		grayLevel = charLevel - 9
	end
	
	-- Check if mob is gray
	if mobLevel <= grayLevel then
		return 0
	end
	
	-- Set the add value (higher for outlands)
	-- TODO: Verify the zone ID's
	local addValue = 45
	if GetZoneID() == 4 then
		addValue = 235
	end
	
	-- Calculate XP
	local XP = 0
	if(charLevel == mobLevel) then
		XP = (charLevel * 5) + addValue
	elseif (charLevel < mobLevel) then
		-- Find max mob level. All mobs higher than (charLevel + 4) are worth the same amount.
		local maxMobLevel = mobLevel
		if mobLevel > charLevel + 4 then 
			maxMobLevel = charLevel + 4 
		end
		XP = ((charLevel * 5 + addValue)) * (1 + (0.05 * (mobLevel - charLevel)))
	elseif (charLevel > mobLevel) then
		local ZD = GetZeroDifference(charLevel)
		XP = ((charLevel * 5 + addValue)) * (1 - ((charLevel - mobLevel) / ZD))
	end
	
	return XP
end
-- Gets the Kills and Quests needed to reach the next level
function GetXToLevel()
	local killsToLevel, questsToLevel

	-- Calculate Mob kills
	local xpPerMob = GetMobAverage()
	if xpPerMob == nil then
		if data['player']['killAverage'] == nil then
			xpPerMob = GetMobXP(currentLevel, currentLevel)
		else
			xpPerMob = data['player']['killAverage']
		end
	end
	
	local restedToLevel = GetRestedToLevel()
	killsToLevel = ceil(((xpTotal - xpCurrent) - restedToLevel) / xpPerMob)
	
	-- Calculate Quest handins
	local xpPerQuest = GetQuestAverage()
	if xpPerQuest == nil and data['player']['questAverage'] == nil then
		questsToLevel = nil
	elseif data['player']['questAverage'] ~= nil then
		questsToLevel = ceil((xpTotal - xpCurrent) / data['player']['questAverage'])
	else
		questsToLevel = ceil((xpTotal - xpCurrent) / xpPerQuest)
	end
	
	return killsToLevel, questsToLevel
end

----------------------
-- Average calculations
----------------------
function GetMobAverage()
	local tableSize = # mobAverageTable
	if tableSize > 0 then
		local total = 0
		local count = 0
		for i,v in ipairs(mobAverageTable) do
			total = total + v
			count = count + 1
		end
		return floor(total / count)
	else
		return nil
	end
end
function AddToMobAverage(xpGained)
	local bufferTable = {}
	bufferTable[1] = xpGained
	local maxStored = 10
	for i,v in ipairs(mobAverageTable) do
		if i >= maxStored then break end
		bufferTable[i+1] = v
	end
	mobAverageTable = bufferTable
end
function ResetMobAverage()
	if # mobAverageTable > 0 then
		mobAverageTable = { GetMobAverage() }
	else
		mobAverageTable = { }
	end
end
function GetQuestAverage()
	if # questAverageTable > 0 then
		local total = 0
		local count = 0
		for i,v in ipairs(questAverageTable) do
			total = total + v
			count = count + 1
		end
		return floor(total / count)
	else
		return nil
	end
end
function AddToQuestAverage(xpGained)
	local bufferTable = {}
	bufferTable[1] = xpGained
	local maxStored = 10
	for i,v in ipairs(questAverageTable) do
		if i >= maxStored then break end
		bufferTable[i+1] = v
	end
	questAverageTable = bufferTable
end
function ResetQuestAverage()
	if # questAverageTable > 0 then
		questAverageTable = { GetQuestAverage() }
	else
		questAverageTable = { }
	end
end

----------------------
-- Event callbacks
----------------------
function MainOnPlayerLogin()
	-- Get XP values
	xpTotal = GetXPMax(UnitLevel("player"))
	xpCurrent = UnitXP("player") or 0
	xpRested = GetXPExhaustion() or 0
	xpRested = (xpRested / 2) -- Divide so it only reflects the additional XP
	
	-- Set chosen locale
	if settings["general"]["useEnglish"] then
		displayLocale = "enUS";
	else
		displayLocale = locale;
	end
	
	-- Initialize GUI config
	ConfigWindow_Init();
	backdropTable = XToLevelDisplay:GetBackdrop();
	
	-- Get X to level
	killsRemaining, questsRemaining = GetXToLevel()
	
	-- Show display
	if settings['averageDisplay']['visible'] then
		XToLevelDisplay:Show();
		UpdateAverageDisplay();
		UpdateAverageBackdrop();
	else
		XToLevelDisplay:Hide();
	end
end
function OnPlayerLevelUp(newLevel)
	-- Set new level and XP values
	currentLevel = newLevel
	xpTotal = GetXPMax(newLevel)
	xpCurrent = 0 -- Set to 0 to prevent old data from being used
	xpRested = GetXPExhaustion() or 0
	xpRested = (xpRested / 2) -- Divide so this only reflects the additional XP
	
	-- Reset averages
	ResetMobAverage()
	ResetQuestAverage()
	
	-- Get kills and quests to level
	killsRemaining, questsRemaining = GetXToLevel()
	
	-- Set display
	UpdateAverageDisplay()
end
function OnChatXPGain(msg)
	-- Update the XP info
	xpTotal = UnitXPMax("player")
	xpCurrent = UnitXP("player")
	xpRested = GetXPExhaustion() or 0
	xpRested = (xpRested / 2) -- Divide so this only reflects the additional XP

	-- Extract the info from the message
	local mobName, xpGained = ExtractKillInfoReg(arg1);
	xpGained = tonumber(xpGained)
	
	-- Add the new XP to the currend XP
	xpCurrent = xpCurrent + xpGained
	
	-- Level reached
	local msg = "n/a"
	if(xpCurrent >= xpTotal) then
		msg = LOCALE_MESSAGES[displayLocale]['LevelReached']
	-- Quest turn-in
	elseif(mobName == nil) then
		-- Calculate the ammount of quest turnins to level
		--  (TotalLevelXP - XPSoFar - GainedXP) / GainedXP
		local quests = ceil((xpTotal - xpCurrent) / xpGained)
		
		-- Create message
		msg = quests .. LOCALE_MESSAGES[displayLocale]['QuestsNeeded']
		
		--  Add to average
		AddToQuestAverage(xpGained)
		data['player']['questAverage'] = GetQuestAverage()
	
	-- Mob kill
	else
		-- Check if the XP gained was rested and remove the rested bonus if needed
		if(xpRested > 0) then
			if(xpRested * 2 < xpGained) then
				xpGained = xpGained - xpRested
			else
				xpGained = xpGained / 2
			end
		end
		
		-- Get How much rested till next level
		local restedToLevel = GetRestedToLevel()
		
		-- Calculate the ammount of mobs needed
		--  (TotalLevelXP - XPSoFar - (RestedBonus)) / GainedXP
		local kills = ceil(((xpTotal - xpCurrent) - restedToLevel) / xpGained)
		
		-- Create Message
		msg = kills .." ".. mobName .. LOCALE_MESSAGES[displayLocale]['KillsNeeded']
		
		-- Add to average
		AddToMobAverage(xpGained)
		data['player']['killAverage'] = GetMobAverage()
	end
		
	-- Send Messages
	ShowPlayerMessages(msg);
	
	-- Update average display and per character values
	killsRemaining, questsRemaining = GetXToLevel()
	UpdateAverageDisplay()
end

----------------------
-- Event handler
----------------------
function MainOnEvent()
	if (event == "PLAYER_LOGIN") then
		MainOnPlayerLogin()
	elseif (event == "PLAYER_LEVEL_UP") then
		OnPlayerLevelUp(arg1)
	elseif (event == "CHAT_MSG_COMBAT_XP_GAIN") then
		OnChatXPGain(arg1)
	end
end
